{{

.:.:-- TACHYON --:.:.
*


A very fast and very small Forth byte code interpreter for the Propeller chip.
2012 Peter Jakacki



NOTE: This is in pre-alpha stage so it's still very messy and the text interpreter is not functional yet
but you can just copy and paste (or download as text) straight into the Spin tool or BST etc as
only the plain text will be copied and it will be ready to compile.

Features:
Low level words are written in PASM and accessed by the
  	Forth run-time interpreter as single byte codes.
  	Byte codes are read from hub RAM and directly address PASM code in cog


Support for LMM operations planned
Interpreted byte code definitions are referenced either as:
 	2 bytes - Call opcode + byte index into bytecode table
  	2 bytes - Call opcode + relative offset back (-255)
  	3 BYTES 16-bit + WCALL token = 3 bytes
All literals and strings are byte aligned
Fast I/O bit-bashing support
Flexible SPI or I2C PASM code support words in kernel
  	Constuct fast serial drivers with minimal code


Holds Forth headers in EEPROM or SD storage
  	Searches the dictionary using rapid index key searching by first character
  	No hub RAM is used by headers
  	Even 32K EEPROMs can be used if the area is in RAM is normally rewritten (i.e. video memory)
  	Option to hold additional information per defintion such as stack usage and description

Kernel compiled in standard manner via Spin tools so other Spin objects can be combined

Three stacks in COG RAM: Data, Return, and Loop
  	Access loop indices outside of definitions allows efficient factoring
  	Avoids manipulation and corruption of return stack
  	Static stack arrays for direct addressing of stack items
  	Intrinsically safe stack overflow and underflow - optional error reporting

Empty loops can execute in 550ns to 875ns (absolute worst case)
Two to one stack operations ( + * AND etc) inc opcode fetch take 900ns to 1.087us (absolute worse case)


The header file is compiled in this file and loaded in but only needs to use it's location
in EEPROM as the image in RAM can be reclaimed for user code or video memory etc.

Headers are in EEPROM automatically when source is compiled in Spin
}}

CON
  _clkmode = xtal1 + pll16x	' <---------------- change to suit
  _xinfreq = 5_000_000		' <---------------- change to suit your crystal

  baud	= 460800  ' test to 2_000_000 (so far)

  scl   = 28
  sda   = 29
  txd   = 30
  rxd   = 31

OBJ

coms : "HS-SerialRx"	' Mega-baud serial receive driver (also transmits if necessary in half-duplex)


PUB Start

  cognew(@RESET, @MAIN)
  word[@rxbufptr] := coms.start(rxd, baud) ' just run slow at 230.4k baud for testing
  ' this Spin cog now terminates and is free for reuse


DAT ' Forth PASM kernel loaded into cog

                        org	0
{
Byte tokens directly address code in the first 256 longs of the cog
Rather than a jump table most functions are short or cascaded to optimize COG memory
Larger fragments of code jump to the second half of the cog's memory.
As a result of not using a jump table (there's not enough memory) there are gaps
in the bytecode values and not all values are usable.

The formatted source has bytecode labels as bold white on red background.
}
'
RESET			jmp	#TACHYON	' As a result of being at location 0 this bytecode = 0

STOP			mov	X,#1
 			cogstop	X
' ( flg -- ) Exit if flg is true
IFEXIT			call	#POPX
 			tjnz	X,#EXIT
 			jmp     #doNEXT
' ( flg -- ) Exit if flg is false (or zero)  Used in place of IF......THEN EXIT as false would just end up exiting
ZEXIT 			call    #POPX
 			tjnz	X,#doNEXT
EXIT			call	#_RPOPIP                  'Pop top of return stack (discard) then cascade into NOP

_NOP 			jmp	#doNEXT

'*************** STACK OPERATIONS ***************

DROP3 			call    #POPX

DROP2			call    #POPX

DROP			call    #POPX
 			jmp     #doNEXT
QDUP			tjz     tos,#doNext

' DUP ( n1 - n1 n1 ) Duplicate the top item on the stack
DUP 			mov     X,tos                   ' Read directly from the top of the data stack
 			jmp     #PUSHX                  ' Push X onto the data stack and doNEXT
' 2DUP ( n1 n2 -- n1 n2 n1 n2 ) Duplicate the top two items on the stack (equiv: OVER OVER)
DUP2 			mov     X,tos+1
 			call    #_PUSHX
OVER			mov     X,tos+1                 'read second data item and push
 			jmp     #PUSHX
THIRD			mov     X,tos+2                 ' read third data item
 			jmp     #PUSHX
' 4TH ( n1 n2 n3 n4 -- n1 n2 n3 n4 n1 ) Copy the 4th item onto the stack
FOURTH			mov     X,tos+3
 			jmp     #PUSHX
' NIP ( n1 n2 -- n2 ) Drop the second item (equiv: SWAP DROP)
NIP			mov     tos+1,tos               'replace second item with top and drop
 			jmp     #DROP
' SWAP ( n1 n2 -- n2 n1 ) Swap the top two items
SWAP			mov     X,tos+1
 			mov     tos+1,tos
 			jmp     #PUTX
' rot   ( a b c -- b c a )
ROT			mov	X,tos+2
 			mov	tos+2,tos+1
  			mov	tos+1,tos
' Replace tos with X as new value
PUTX			mov	tos,X
 			jmp	#doNEXT


'********************* ARITHMETIC *******************

PLUS			movi    _TWOOP,#%100000_001    ' + ( n1 n2 -- n3 )
 			jmp     #TWOOP
MINUS			movi    _TWOOP,#%100001_001
 			jmp     #TWOOP
INC			add     tos,#1                  '1+
 			jmp     #doNEXT
DEC 			sub     tos,#1                  '1-
 			jmp     #doNEXT

DIVIDE			mov     X,tos
 			jmp     #_DIVIDE

MULTIPLY			call    #_UMMUL                ' *
 			jmp     #DROP

UMMUL			call    #_UMMUL                ' UM*
 			jmp     #doNEXT

NEGATE			mov     X,tos
 			mov     tos,#0
 			sub     tos,X
 			jmp     #doNEXT


'****************** BOOLEAN ******************

_NOT                    xor     tos,_minus1
                        jmp     #doNEXT
_AND                    movi    _TWOOP,#%011000_001    ' AND ( n1 n2 -- n3 )
                        jmp     #TWOOP                  'discard top of stack and execute modified PASM
_OR                     movi    _TWOOP,#%011010_001
                        jmp     #TWOOP
_XOR                    movi    _TWOOP,#%011011_001
                        jmp     #TWOOP
_SHR                    movi    _TWOOP,#%001010_001
                        jmp     #TWOOP
_SHL                    movi    _TWOOP,#%001011_001
                        jmp     #TWOOP
' 2/ ( n1 -- n1 ) shift n1 right one bit (equiv to divide by 2)
_SHR1                   shr     tos,#1
                        jmp     #doNEXT
' 2* ( n1 -- n2 ) shift n1 left one bit (equiv to multiply by 2)
_SHL1                   shl     tos,#1
                        jmp     #doNEXT

' REV ( n1 bits -- n2 ) Reverse LSBs of n1 and zero-extend
_REV                    movi    _TWOOP,#%001111_001
                        jmp     #TWOOP

' MASK ( bitpos -- bitmask )
MASK                    mov     X,tos
                        mov     tos,#1
                        shl     tos,X
                        jmp     #doNEXT

'*************** COMPARISON ******************

' 0= ( n1 -- flg ) true if n1 equals 0
ZEQ			or	tos,#0 wz
SETZ 			neg	tos, #1
 		if_nz	mov	tos,#0 	'force true to false
 			jmp	#doNEXT

' = ( n1 n2 -- flg ) true if n1 is equal to n2
EQ 			cmp	tos,tos+1  wz
 			neg	tos+1,#1
		if_nz	mov	tos+1,#0
			jmp     #DROP

' > ( n1 n2 -- flg ) true if n1 > n2
GT			cmps	tos+1,tos  wz,wc
                        call	#POPX
        if_a            neg	tos,#1
        if_be           mov	tos,#0
                        jmp	#doNEXT


'***************** MEMORY *******************

CFETCH                  rdbyte  tos,tos
                        jmp     #doNEXT

WFETCH                  rdword  tos,tos
                        jmp     #doNEXT

FETCH                   rdlong  tos,tos
                        jmp     #doNEXT
' ( n addr -- )
CPLUSST                 rdbyte  X,tos           ' read in word from adress
                        add     tos+1,X         ' add to contents of address - cascade
CSTORE                  wrbyte  tos+1,tos       ' write the byte using address on the tos
                        jmp     #DROP2

' C@++  ( addr -- addr+1 ch ) fetch char and increment address
CFETCHINC               mov     X,tos                   ' dup the address
                        call    #_PUSHX
                        add     tos+1,#1                ' inc the backup address
                        jmp     #CFETCH                 ' fetch at the current byte

WPLUSST                 rdword  X,tos           ' read in word from adress
                        add     tos+1,X
WSTORE                  wrword  tos+1,tos
                        jmp     #DROP2

PLUSST                  rdlong  X,tos           ' read in long from adress
                        add     tos+1,X
STORE                   wrlong  tos+1,tos
                        jmp     #DROP2

CMOVE                   jmp     #_CMOVE                 'CMOVE

' SET ( mask addr -- ) Set bit in hub byte
SET                     movi    MEMINS,#%011010_001     ' or
                        jmp     #MEMOP
CLR                     movi    MEMINS,#%011001_001     ' andn
                        jmp     #MEMOP
' BIT? ( mask caddr -- flg ) Test single bit of byte in memory
BTEST                   rdbyte  X,tos
                        and     X,tos+1 wz
                        neg     tos+1, #1
              if_z      mov     tos+1,#0
                        jmp     #DROP

'**************** LITERALS ******************

' Push a 32-bit literal onto the datastack by reading in the next 4 bytes (non-aligned)
PUSH4                   call    #ACCBYTE                ' read the next byte @IP++ and shift accumulate
' Push a 24-bit literal onto the datastack by reading in the next 3 bytes (non-aligned)
PUSH3                   call    #ACCBYTE
' Push a 16-bit literal onto the datastack by reading in the next 2 bytes (non-aligned)
PUSH2                   call    #ACCBYTE
' Push an 8-bit literal onto the datastack by reading in the next byte
_BYTE
PUSH1                   call    #ACCBYTE
                        call    #_PUSHACC               ' Push the accumulator onto the stack then zero it
                        jmp     #doNEXT


' Variables start with this single byte code which returns with the address of the long aligned variable following
VARL                    mov     X,IP
                        add     X,#3	' force long alignment
                        andn    X,#3
PUSHX_EXIT              call    #_PUSHX	' push address of variable
                        jmp     #EXIT
' Byte variables start with this single byte code which returns with the address of the byte variable following
VARB                    mov     X,IP
                        jmp     #PUSHX_EXIT

' ******************* FAST CONSTANTS ********************

' Push a preset literal onto the stack using just one bytecode
'
_FALSE
_0                      mov     X,#0
                        jmp     #PUSHX
_1                      mov     X,#1
                        jmp     #PUSHX
_2                      mov     X,#2
                        jmp     #PUSHX
_3                      mov     X,#3
                        jmp     #PUSHX
_4                      mov     X,#4
                        jmp     #PUSHX
_8                      mov     X,#8
                        jmp     #PUSHX
_0F                     mov     X,#$0F
                        jmp     #PUSHX
_0FF                    mov     X,#$FF
                        jmp     #PUSHX
_TRUE
MINUS1                  neg     X,#1
                        jmp     #PUSHX
BL                      mov     X,#$20
                        jmp     #PUSHX

'***************** I/O CONTROL ******************

' P@ ( -- n1 ) Read the input port A (assume it is always A for Prop 1)
PFETCH                  mov     X,INA
                        jmp     #PUSHX
' P! ( n1 -- ) Store n1 to the output port A
PSTORE                  mov     OUTA,tos
                        jmp     #DROP
' OUTSET ( mask -- ) Set multiple bits on the output
OUTSET                  or      OUTA,tos                'PSET ( mask -- ) \ Or mask to OUTA and make sure it's an output
                        jmp     #OUTPUTS
' OUTCLR ( mask -- ) Clear multiple bits on the output
OUTCLR                  andn    OUTA,tos                'PCLR ( mask -- )

' OUTPUTS ( mask -- ) Set selected port pins to outputs
OUTPUTS			or      DIRA,tos                'PDSET ( mask -- ) \ OR mask to DIRA
 			jmp     #DROP
' INPUTS ( mask -- ) Set selected port pins to inputs
INPUTS			andn    DIRA,tos
 			jmp     #DROP
' px ( t/f n1 -- ) set pin # n1 to h - true or l false  (method borrowed from PropForth)
PX                      or      tos+1,#0 wz
                        muxnz   OUTA,tos
                        jmp     #DROP2

' CLKBIT ( outmask clkmask dat -- outmask clkmask dat/2 ) Clock to low then high the lsb of dat and shift
CLKBIT                  call    #_CLKBIT
                        jmp     #doNEXT

' CLKBITS ( outmask clkmask dat bits -- outmask clkmask dat/2 )
CLKBITS                 jmp     #_CLKBITS

'***************** COG ACCESS ******************

' COG@ ( addr -- long ) Fetch a long from cog memory
COGFETCH		movs	_cf,tos
_cf 			mov	tos,0_0
 			jmp	#doNEXT

' COG! ( long addr -- ) Store a long to cog memory
COGSTORE			movd    _cd,tos
_cd			mov     0_0,tos+1
 			jmp     #DROP2

' HUBOP ( val op -- )
_HUBOP                  hubop   tos,tos+1
                        jmp     #DROP2

' STACKS ( -- cog_addr ) push address of internal stacks in cog memory
STACKS			mov	X,#tos
			jmp	#PUSHX
' Loop control words such as J K LEAVE etc implemented
' LSTACK ( -- cog_addr ) push address of the loop stack in cog memory
LSTACK			mov	X,#loopstk
			jmp	#PUSHX

'PASM ( code -- ) Load and execute PASM instruction


'***************** BRANCH & LOOP *********************

' Perform a call to kernel bytecode via the kbctbl using the following inline byte as an index into that table
XCALL                   call    #SETUPIP                ' read offset in table
                        shl     X,#2                    ' offset into longs in hub RAM
                        add     X,kbcptr                '
                        rdword  IP,X
                        jmp     #doNEXT
' Call a local defintion using 8-bit relative = 0 to -255
RCALL                   call    #SETUPIP                ' read next byte into X and save IP in return stack
                        jmp     #JMPBACK
' Call any bytecode definition using 16-bit relative addresses
WCALL                   call    #GETBYTE
                        jmp     #_WCALL                 ' Call any byte code definition using a 16-bit address

' ( n1 n2 -- ) Compare two values and jump if equal. Equivalent to <> IF but clearer in Spin tool
CMPJEQ                  cmp     tos,tos+1 wz,wc
                        call    #POPX
                        call    #POPX
              if_ne     jmp     #SKIP

' Jump forward by reading the next byte inline and adding that to the IP (at that point)
JUMP
_ELSE                   call    #GETBYTE
                        jmp     #JMPFWD

' If flg is zero than jump forward by reading the next byte inline and adding that to the IP (at that point)
' IF R( flg -- )
JZ
_IF                     movi    zBRINS,#%100000_001     ' add
                        jmp     #zBRANCH
JZBACK
_UNTIL                  movi    zBRINS,#%100001_001     ' sub
                        jmp     #zBRANCH
JUMPBACK
_AGAIN                  call    #GETBYTE
                        jmp     #JMPBACK
' ADO = BOUNDS DO - just a quick and direct way as BOUNDS is most often never used elsewhere
ADO                     mov     X,tos+1
                        add     tos+1,tos
                        mov     tos,X
DO                      call    #_PUSHLP                  ' PUSH index onto loop stack
                        mov     tos,tos+1
                        call    #_PUSHLP                  ' push limit onto loop stack
                        jmp     #DROP2

LOOP			add     loopstk+1,#1 wc             ' increment index,
 			jmp     #LOOPCHK
' (+loop) ( n1 -- ) adds n1 to the loop index and branches back if not equal to the loop limit
PLOOP			add     loopstk+1,tos wc
 			call    #POPX
LOOPCHK			cmps     loopstk,loopstk+1 wz,wc
 		if_ne	jmp     #_AGAIN
 			call    #_LPPOP
SKIP1			call    #_LPPOP
SKIP			add     IP,#1
 			jmp     #doNEXT


' FOR ( count -- ) Setup FOR...NEXT loop for count
FOR                     call    #_PUSHLP
                        jmp     #DROP
' NEXT ( -- ) Decrement count (on loop stack) and loop until 0, then pop loop stack
forNEXT                 djnz    loopstk,#JUMPBACK
                        jmp     #SKIP1

' >R ( n -- ) Push n onto the return stack
PUSHR                   mov     R0,tos
                        call    #_PUSHR
                        jmp     #DROP
' R> ( -- n ) Pop n from the return stack
RPOP                    call    #_RPOP                  ' Pop return stack into R and X
                        jmp     #PUSHX                  ' Push X onto the data stack as tos

' >L ( n -- ) Push n onto the loop stack
PUSHL                   call    #_PUSHLP
                        jmp     #DROP
' L> ( -- n ) Pop n from the loop stack
LPOP                    call    #_LPPOP                  ' Pop return stack into R and X
                        jmp     #PUSHX                  ' Push X onto the data stack as tos
{
LEAVE                   mov     loopstk+1,loopstk
                        jmp     #doNEXT
}
' I ( -- index ) Read current loop index of DO..LOOP
I 			mov	X,loopstk+1
 			jmp	#PUSHX

EMIT 			mov	X,tos
 			call	#transmit
 			jmp	#DROP

' Registers can be used just like variables and the interpreted kernel uses some for itself
' 64 bytes are reserved which can be accessed as bytes/words/longs depending upon
' the alignment.

' ( -- addr ) Read the next inline byte and return with the register byte address
REG			call	#GETBYTE
 			call	#_PUSHX
' ( index -- addr )
ATREG			add	tos,regptr
 			jmp	#doNEXT

' WAIT ( n -- ) Wait n clock cycles
WAIT                    add     tos,cnt
                        waitcnt tos,tos
                        jmp     #DROP
{
' WAITCNT ( n -- n )
WAITCNTS			waitcnt	tos,tos
 			jmp	#doNEXT
}
' Print a null terminated inline string
PRTSTR                  jmp     #_PRTSTR
'
' Compare a null-terminated source string with a dictionary string which is 8th bit terminated.
' This will always force a mismatch after which one is checked for a null while the other is checked
' for the 8th bit and if verified then a match has been found.
' The dict pointer is advanced to point to the end of the dict string on the 8th bit termination which
' is the attribute byte as in: byte "CMPSTR",$80,CMPSTR
'
'
' CMPSTR ( src dict -- src dict+ flg ) Compare strings at these two addresses
CMPSTR			jmp	#_CMPSTR

' The read and write index is stored as two bytes preceding the buffer, read this as a word (faster)
' BKEY ( buffer -- ch ) ' byte size buffer is preceded with a read index, go and read the next character
' assume the buffer is 256 bytes long
BKEY
			mov	X,tos
			sub	X,#2	' X Point to read (& write index)
			rdword	Z,X
			mov	Y,Z
			shr	Y,#8	' get read index Y
			and	Z,#$0ff	' mask write index Z
			cmp	Y,Z wz	' compare indicies
		if_z	mov	tos,#0	' Return with a null value
		if_z	jmp	#doNEXT
			add	tos,Z	' tos = read pointer
			rdbyte	tos,tos
			or	tos,#$100	' and mark as valid (in case of nulls)
			add	Z,#1
			and	Z,#$0ff	' wrap-around
			wrbyte	Z,X	' update read index
			jmp	#doNEXT





{{*********************************** INTERNAL COG ROUTINES *********************************
Code from here at cog address $100 on is not meant to be indexed
by byte codes so therefore can be placed anywhere }}



regptr		long @registers+$10
kbcptr		long @kbctbl+$10                  ' used by XCALL
fthptr		long @MAIN+$10

' Reset Entry
TACHYON
                       ' mov     tos,#0
                       ' mov     retstk,#0
                       ' mov     loopstk,#0
 			mov	IP,fthptr 'PAR
                       ' rdlong  IP,IP                   '@00 read 1st long in hub memory to init IP


' ****************************** The runtime bytecode interpreter **********************
'
' Fetch the next byte code instruction in hub RAM pointed to by the instruction pointer IP
'
doNEXT
 			rdbyte	token,IP	'read byte code instruction
 			add	IP,#1	'advance IP to next byte token
 			jmp	token	'execute the code


' common operations for two data items into one
TWOOP                   call    #POPX
_TWOOP                  add     tos,X wz,wc          ' instruction here is modified by caller
'                        muxc    status,#carry	' optional status flag update
'                        muxz    status,#zero
                        jmp     #doNEXT

MEMOP                   rdbyte  X,tos
MEMINS                  andn    X,tos+1
                        wrbyte  X,tos
                        jmp     #DROP2

' Push IP>R, read 16-bit address relative +/- address into IP
_WCALL                 'call    #GETBYTE
                        mov     Y,X
                        call    #SETUPIP
                        shl     X,#8
                        or      X,Y
JMPFWD                  add     IP,X
                        jmp     #doNEXT

JMPBACK                 sub     IP,X
                        jmp     #doNEXT

' Read the next byte as a displacement and branch forward or backwards (BRINS modified by caller)
zBRANCH                 call    #GETBYTE        'read in next byte at IP and inc IP
                        or      tos,#0 wz       'test flag on stack
zBRINS        if_z      add     IP,X            'Adjust IP forward according to flag
                        jmp     #DROP           'discard flag



'**************************** LITERALS *********************

ACCBYTE 			call	#GETBYTE	' Build a literal by reading in another byte
 			shl	ACC,#8	' merge it into the "accumulator" byte by byte
 			or	ACC,X
ACCBYTE_ret 		ret

GETBYTE			rdbyte	X,IP	' Simply read a byte (non-code) and advance the IP
  			add	IP,#1
GETBYTE_ret 		ret

PUSHX			call	#_PUSHX	' Push the internal X register onto the datastack
  			jmp	#doNEXT


'*************************** ARITHMETIC ************************

_DIVIDE                 'mov     X,tos                        ' GET A SIGN

 			shl	tos,#1 wc,NR
 			negc	tos,tos
 			mov	Y,tos
 			shl	tos+1,#1 wc,NR
 			negc	tos+1,tos+1
 			xor	X,Y 	' Compute result sign
' *** Now on to unsigned division
 			mov	R0,#32
 			mov	R1,#0                          ' Do unsigned division
Dloop			shl	tos,#1   wc
 			rcl	R1,#1
 			cmpsub	R1,tos+1  wc,wr
 			rcl	R2,#1
 			djnz	R0,#Dloop
' *** End of unsigned division
 			shl	X,#1 wc,NR                 ' Apply result sign
 			negc	tos,R2
 			jmp	#doNEXT

' um* ( u1 u2 -- u1*u2L u1*u2H ) \ unsigned 32bit * 32bit -- 64bit result
_UMMUL
 			cmp	tos+1,tos wc    ' try to go faster by using the lower number as the test factors
  		if_nc	mov	R0,tos+1
 		if_nc	mov	R2,tos
 		if_c	mov	R0,tos           ' R0R1 = u2
 		if_c	mov	R2,tos+1         ' R2R3 = u1
 			mov	R1,#0
''''''			mov	R3,#0
 			mov	tos,#0           ' zero result
 			mov	tos+1,#0
UMMULLP 			shr	R2,#1 wz,wc     ' test next bit of u1
 		if_nc	jmp	#UMMUL1
 			add	tos+1,R0 wc       ' add in shifted u2
 			addx	tos,R1           ' carry into upper long
UMMUL1			add	R0,R0 wc          ' shift u2 left
 			addx	R1,R1            ' carry into 64-bits
 		if_nz	jmp	#UMMULLP       ' exhausted u1?
_UMMUL_ret		ret


' Print inline string
_PRTSTR                 rdbyte  X,IP
                        add     IP,#1
                        or      X,#0 wz
              if_z      jmp     #doNEXT
                        call    #transmit
                        jmp     #_PRTSTR

_CMPSTR 			call	#_PUSHX	' make room for a flag
 			mov	X,tos+2	' X = source
cmpstrlp 			rdbyte	R0,X	' read in a character from the source
 			rdbyte	R1,tos+1	' read in from the dictionary
			cmp	R1,R0 wz	' are they the same?
 		if_nz	jmp	#nomatch
			add	X,#1	' so far, so good, try next
			add	tos+1,#1	' updates the dict pointer on the stack too
			jmp	#cmpstrlp	' keep at it
nomatch			cmp	R0,#0 wz	' was the src null terminated?
		if_z	xor	R1,#$80 wz	' set z flag if dict 8th bit set
			jmp	#SETZ


' CMOVE ( src dst cnt -- )
_CMOVE			rdbyte  X,tos+2         ' read source byte
 			add     tos+2,#1
 			wrbyte  X,tos+1         ' write destination byte
 			add     tos+1,#1
 			djnz    tos,_CMOVE
 			jmp     #DROP3

' CLKBIT ( outmask clkmask data -- outmask clkmask data ) NOTE: doesn't set output direction
_CLKBIT			andn    OUTA,tos+1      ' clock low
			test	tos+2,INA wz	' read data from line
 			shr     tos,#1 wc
   			muxz	tos,bit31	' read data goes into msb
			muxc    OUTA,tos+2
 			or      OUTA,tos+1      ' clock hi
_CLKBIT_ret		ret

' CLKBITS ( outmask clkmask data cnt -- outmask clkmask data ) Clocks bits at 2.85MHz
_CLKBITS 			call    #POPX
 			or      DIRA,tos+1      ' make clock an output
 			or      DIRA,tos+2      ' make out an output
CLKBITlp			call    #_CLKBIT
 			djnz    X,#CLKBITlp
 			jmp     #DROP3

bit31		long	$8000_0000
' ************************* Internal fixed stack method *******************
'
' As well as the data and return stack, a loop stack is also employed
' The return stack is only used for return addresses

{ The stacks are longer than necessary for testing purposes - will optimize later }

' Pop the data stack using fixed size stack in COG memory (fast for other operations)
POPX 			mov	X,tos	' pop old tos into X (temporary)
 			mov	tos,tos+1
 			mov	tos+1,tos+2
 			mov	tos+2,tos+3
 			mov	tos+3,tos+4
  			mov	tos+4,tos+5
 			mov	tos+5,tos+6
 			mov	tos+6,tos+7
 			mov	tos+7,tos+8
 			mov	tos+8,tos+9
 			mov	tos+9,tos+10
 			mov	tos+10,tos+11
 			mov	tos+11,#0	' zero fill for error checking
POPX_ret 			ret

_PUSHACC			mov	X,ACC
 			mov	ACC,#0
_PUSHX			tjnz	tos+11,#ERROR	' check for overflow - zeros are don't care
 			mov	tos+11,tos+10
 			mov	tos+10,tos+9
 			mov	tos+9,tos+8
 			mov	tos+8,tos+7
 			mov	tos+7,tos+6
 			mov	tos+6,tos+5
 			mov	tos+5,tos+4
 			mov	tos+4,tos+3
 			mov	tos+3,tos+2
 			mov	tos+2,tos+1
 			mov	tos+1,tos
 			mov	tos,X	' replace tos with X (DEFAULT)
_PUSHACC_ret
_PUSHX_ret		ret


' LOOP STACK
_PUSHLP  ' Push tos onto the loop stack
                        mov     loopstk+7,loopstk+6
                        mov     loopstk+6,loopstk+5
                        mov     loopstk+5,loopstk+4
                        mov     loopstk+4,loopstk+3
                        mov     loopstk+3,loopstk+2
                        mov     loopstk+2,loopstk+1
                        mov     loopstk+1,loopstk
                        mov     loopstk,tos
_PUSHLP_ret              ret

_LPPOP  ' pop the loop stack into X
                        mov     X,loopstk
                        mov     loopstk,loopstk+1
                        mov     loopstk+1,loopstk+2
                        mov     loopstk+2,loopstk+3
                        mov     loopstk+3,loopstk+4
                        mov     loopstk+4,loopstk+5
                        mov     loopstk+5,loopstk+6
                        mov     loopstk+6,loopstk+7
                        mov     loopstk+7,#0
_LPPOP_ret              ret

'*************************** RETURN STACK ****************************
'
' NOTE: Brute force method for return stack being replaced with indexed methdod
' as return stack items do not need to be directly addressed
' This indexed method does not use movd and movs methods but directly inc/decs the
' source and destination fields of the instruction.
'
SETUPIP			call	#GETBYTE	' read the next byte into X and save the current IP
SAVEIP			mov	R0,IP
_PUSHR			mov	retstk,R0	' save it on the stack (dest modified)
 			add	rpopins,#1	' update source for popping
 			add	_PUSHR,dst1	' update dest for pushing
SETUPIP_ret
SAVEIP_ret
_PUSHR_ret		ret

' RETURN STACK - pop into X
_RPOPIP			call	#_RPOP
 			mov	IP,X
_RPOPIP_ret		ret

_RPOP			sub	rpopins,#1
 			sub	_PUSHR,dst1
rpopins			mov	X,retstk
_RPOP_ret			ret

dst1			long	$200

{ Just a basic error routine to let me know at present if there has been a problem }
ERROR
 			mov     X,#"E"
 			call    #transmit
 			mov     X,#"R"
 			call    #transmit
  			mov     X,#"R"
 			call    #transmit
 			jmp     #STOP

{---------------------------------- CONSOLE OUTPUT -----------------------------}
'
'************************ Serial Transmit ************************
'
transmit			or      outa,txmask
 			or      dira,txmask
 			or      X,stopmask	' fill up with stop bits
 			shl     X,#1	' add in start bit as first bit
 			mov     txcnt,#11	' 8 data bits + 1 start + stop bits
txdat			mov     R0,cnt
 			add     R0,bit_ticks
txbit 			shr     X,#1       wc	' lsb first
 			muxc    outa,txmask	' output bit
 			waitcnt R0,bit_ticks
 			djnz    txcnt,#txbit	'another bit to transmit?
transmit_ret 		ret


bit_ticks		long (80_000_000 / baud )	' set baud rate
txmask		long |<txd
stopmask		long    $FFFFFFFF<<8

txcnt		res 1

'****************** COG VARIABLES *****************
'
_minus1		long -1
_zero		long 0

tos
datastk		long 0[12]
retstk		long 0[11]
loopstk		long 0[8]

token		res 1
IP		res 1	' Instruction pointer
ACC		res 1	' Accumulation register for inline byte-aligned literals
R0		res 1
R1		res 1
R2		res 1
X		res 1	'primary internal working registers
Y		res 1	'
Z		res 1

                        fit 496
{

}
'***************************** HIGH LEVEL FORTH CODE *******************

CON
carry 		= 1
zero		= 2

tibsz 		= 80


DAT

' Allocate storage memory for buffers and other variables

rxbufptr		long 0
TIB    		byte 0[tibsz]		' traditional text input buffer
registers		long 0[32]			'Variables used by kernel + general-purpose


	org  0
' register offsets within "registers". Access as    REG,DELIM   ...  REG,BASE ... etc
'
temp	res 20
addr	res 4
UNUM	res 8
EXECLOC	res 4
TIBPTR	res 2
DELIM	res 2
EOLTIB	res 1
IN	res 1
CHARS	res 1
NUMFLG	res 1
DPL	res 1
FLAGS	res 1
BASE	res 1
BASESAV	res 1
DIGITS	res 1
wordcnt	res 1
wordbuf	res 32
rxbuf	res 2

'

{                VECTOR TABLE for KERNEL BYTECODES                     }

{  Kernel bytecode definitions need to be called and this table makes it easy to do so
  with just a 2 byte call. Extra memory may be allocated for user definitions as well
  The Spin compiler requires longs whereas we only need 16-bit words but this will do
  at present. }


    org 0		' ensure references can be reduced to a single byte index
kbctbl
xEXEC		long @registers+EXECLOC+$10
xNEQ		long @NEQ+$10
xZNE		long @ZNE+$10
xLT		long @LT+$10
xZLT		long @ZLT+$10
xZGT		long @ZGT+$10
xULT		long @ULT+$10
xWITHIN		long @WITHIN+$10
xMS		long @ms+$10
xLEAVE		long @LEAVE+$10
xCOMMENT		long @COMMENT+$10
xPRTHEX		long @PRTHEX+$10
xPRTBYTE		long @PRTBYTE+$10
xPRTWORD		long @PRTWORD+$10
xPRTLONG		long @PRTLONG+$10
xPRTSTK		long @PRTSTK+$10
xPRTSTKS		long @PRTSTKS+$10
xDEBUG		long @DEBUG+$10
xDUMP		long @DUMP+$10
xCOGDUMP		long @COGDUMP+$10
xCR		long @CR+$10
xOK		long @OK+$10
xBIN		long @BIN+$10
xDECIMAL		long @DECIMAL+$10
xHEX		long @HEX+$10
xKEY		long @KEY+$10
xQEMIT		long @QEMIT+$10
xTOUPPER		long @TOUPPER+$10
xPUTCHAR		long @PUTCHAR+$10
xdoCHAR		long @doCHAR+$10
xGETWORD		long @GETWORD+$10

xFINDSTR		long @FINDSTR+$10
xEXECUTE		long @EXECUTE+$10
xVER		long @VER+$10
xFILL		long @FILL+$10
xERASE		long @ERASE+$10
xCMOVE		long @CMOVE+$10
xTerminal		long @Terminal+$10


' NOTE: this table is limited to 256 entires
             'long 0[128]                       ' leave room for more table entries for XCALLs (0=empty)

{ Moving into the Forth extensions written directly in bytecode }


{ debug print routines - also used by DUMP etc }

PRTHEX  ' ( n -- ) print n (0..$0F) as a hex character
        byte  _0F,_AND
        byte  PUSH1,$30,PLUS
        byte  DUP,PUSH1,$39,GT,_IF,@PRTCH-@PH1
PH1     byte  PUSH1,7,PLUS                      'Adjust for A..F
PRTCH   byte  EMIT,EXIT

PRTBYTE
        byte  DUP,_4,_SHR
        byte  XCALL,xPRTHEX,XCALL,xPRTHEX,EXIT
        byte  'RCALL,@PB1-@PRTHEX
PB1     byte  'RCALL,@PB2-@PRTHEX
PB2     byte  EXIT

PRTWORD
        byte  DUP,_8,_SHR
        byte  RCALL,@PW1-@PRTBYTE
PW1     byte  RCALL,@PW2-@PRTBYTE
PW2     byte  EXIT

PRTLONG
        byte  DUP,PUSH1,16,_SHR
        byte  RCALL,@PRL1-@PRTWORD
PRL1    byte  RCALL,@PRL2-@PRTWORD
PRL2    byte  EXIT

DUMP  	' ( addr cnt -- ) Hex dump of hub RAM - (if CFETCH is vectored then other memory can be accessed)
 	byte  ADO
DML	byte  XCALL,xCR
 	byte  I,XCALL,xPRTWORD
 	byte  PRTSTR,": ",0
 	byte  I,PUSH1,$10,ADO
DM0	byte  I,CFETCH,XCALL,xPRTBYTE
  	byte  PUSH1,$20,EMIT,LOOP,@DM2-@DM0
DM2	byte  PUSH1,$10,PLOOP,@DM3-@DML
DM3	byte  EXIT

COGDUMP	' ( addr cnt -- ) Dump cog memory, but try to minimize stack usage
	byte  REG,0,STORE,REG,4,STORE
cdmlp	byte  XCALL,xCR,REG,4,FETCH,XCALL,xPRTWORD,PRTSTR,": ",0
 	byte	REG,4,FETCH,COGFETCH,XCALL,xPRTLONG
	byte	_1,REG,4,WPLUSST,MINUS1,REG,0,WPLUSST
	byte	REG,0,FETCH,ZEQ,_UNTIL,@cdm1-@cdmlp
cdm1 	byte	EXIT


' Print out the top four numbers of the datastack
PRTSTK	byte  PRTSTR,$0D,$0A,"STACK: ",0
 	byte  FOURTH,XCALL,xPRTLONG,BL,EMIT
 	byte  THIRD,XCALL,xPRTLONG,BL,EMIT
 	byte  OVER,XCALL,xPRTLONG,BL,EMIT
 	byte  DUP,XCALL,xPRTLONG,BL,EMIT
 	byte  EXIT

PRTSTKS	' Print stacks but avoid cluttering with data from debug routines
	byte	XCALL,xPRTSTK
	byte	STACKS,PUSH1,12,PLUS,PUSH1,11
	byte	PRTSTR,$0D,$0A,"RETURN STACK ",0
	byte	XCALL,xCOGDUMP
	byte	STACKS,PUSH1,23,PLUS,PUSH1,8
	byte	PRTSTR,$0D,$0A,"LOOP STACK ",0
	byte	XCALL,xCOGDUMP
	byte	EXIT

' Print the stack(s) and dump the registers
DEBUG	byte	XCALL,xPRTSTK,XCALL,xPRTSTKS
 	byte	PRTSTR,$0D,$0A,"REGISTERS",0
 	byte	REG,0,PUSH1,$80,XCALL,xDUMP,EXIT


' 0<> ( n1 -- flg ) true if n1 is not equal to 0
ZNE	byte  ZEQ,_NOT,EXIT

' <> ( n1 n2 -- flg ) true if n1 is not equal to n2
NEQ	byte  EQ,_NOT,EXIT

' 0>
ZGT	byte  _0,GT,EXIT

' 0<
ZLT	byte  _0,XCALL,xLT,EXIT

LT	byte  SWAP,GT,EXIT

' U<
ULT	byte  DUP2,_XOR,XCALL,xZLT,_IF,04
	byte  NIP,XCALL,xZLT,EXIT
 	byte  MINUS,XCALL,xZLT,EXIT

' ( n lo hi -- flg ) true if n is within range of low and high inclusive
WITHIN	byte	INC,OVER,MINUS,PUSHR
	byte	MINUS,RPOP,XCALL,xULT
WT1	byte	XCALL,xZNE,EXIT

'	\	( -- )
'	Ignore following text till the end of line.
'	IMMED
COMMENT	byte	REG,EOLTIB,CFETCH,REG,IN,CSTORE,EXIT

LEAVE	byte	LSTACK,COGFETCH,LSTACK,INC,COGSTORE,EXIT	' set index to the same as limit

ERASE	byte	_0 'FILL
' ( addr cnt fillch -- )
FILL	byte	ROT,ROT,ADO,DUP,I,CSTORE,LOOP,5,DROP,EXIT

' ms ( n -- ) Wait for n milliseconds
ms	 byte  PUSH3,$01,$38,$80,MULTIPLY,WAIT,EXIT

' IIC! ( byte -- ) Write byte to IIC bus
IICST	byte	PUSH1,scl,MASK,PUSH1,sda,MASK,ROT
 	byte	_8,CLKBITS
	byte	EXIT

KEY	byte	REG,rxbuf,WFETCH,WFETCH,BKEY,QDUP,_UNTIL,@ky1-@KEY
ky1	byte	_0FF,_AND,EXIT

'***********************************************************




' change the default number bases
BIN     byte  _2,JUMP,@SetBase-@DECIMAL

DECIMAL byte  PUSH1,10,JUMP,@SetBase-@HEX

HEX     byte  PUSH1,16
SetBase byte  REG,BASE,CSTORE,EXIT


CR      byte  PUSH1,$0D,EMIT,PUSH1,$0A,EMIT,EXIT


' PROMPT
OK     byte  PRTSTR," ok",$0D,$0A,0,EXIT



' ?EMIT	,( ch -- ch )
QEMIT   byte _1,REG,flags,BTEST,ZEXIT,DUP,EMIT,EXIT


' UPPER  ( str1 --  ) Convert lower-case letters to upper-case
TULP    byte  INC
TOUPPER
        byte  DUP,CFETCH,QDUP,_IF,@TUX-@TU1                  ' end of string?
TU1     byte  _BYTE,"a",_BYTE,"z",XCALL,xWITHIN
        byte  _UNTIL,@TU2-@TULP
TU2     byte  _BYTE,-$20,OVER,CPLUSST,_AGAIN,@TUX-@TULP
TUX     byte  DROP,EXIT


EXECUTE	' ( bytecode1 bytecode2 -- )
	byte	REG,EXECLOC+1,CSTORE,REG,EXECLOC,CSTORE
 	byte	PUSH1,EXIT,REG,EXECLOC+2,CSTORE,xCALL,xEXEC
 	byte	EXIT


{
Replacing traditional parse function with realtime stream parsing
Each word is acted upon when a delimiter is encountered and this also allows for
interactive error checking and even autocompletion.
}

' ( ch -- ) write a character into the next free position in the word buffer
PUTCHAR	byte	REG,wordcnt,DUP,CFETCH,SWAP,INC,PLUS,CSTORE,EXIT

' As characters are accepted from the input stream, checks need to be made for delimiters,
' editing commands etc.
doCHAR	' ( char -- flg ) Process char into wordbuf and flag true if all done
	byte	DUP,ZEXIT
	byte	PUSH1,$09,OVER,EQ,_IF,02,DROP,BL		' substitute a tab with a space
	byte	PUSH1,$0D,OVER,EQ,_IF,@dc1-@cr1
cr1	byte	DROP,_TRUE,EXIT		' CR - Return & indicate completion
dc1	byte	PUSH1,$08,OVER,EQ,_IF,@tk2-@bksp1		' BKSP - null out last char
bksp1	byte	DROP,REG,wordcnt,CFETCH,_IF,@tk2-@bksp2		' don't backspace if empty word
bksp2 	byte	MINUS1,REG,wordcnt,CPLUSST,_0,XCALL,xPUTCHAR	' null previous char
	byte	MINUS1,REG,wordcnt,CPLUSST		' maintain correct index
	byte	_FALSE,EXIT			'
tk2	byte	PUSH1,$0A,OVER,EQ,_IF,03,DROP,_FALSE,EXIT		' discard LFs
	byte	REG,DELIM,CFETCH,OVER,EQ,_IF,@tk5-@adelim		' delimiter?
adelim 	byte	DROP,REG,wordcnt,CFETCH,EXIT		' true if trailing delimiter - all done
tk5	' otherwise build text in wordbuf - null terminated with a preceding count .....
	byte	XCALL,xPUTCHAR			' put a character into the word buffer
	byte	_1,REG,wordcnt,CPLUSST		' save char and advance ptr -
	byte	_FALSE,EXIT

' Build a delimited word and return immediately upon a valid delimiter
GETWORD	' ( -- ) Build a text word from character input into wordbuf for wordcnt
	byte	REG,wordcnt,PUSH1,33,XCALL,xERASE		'Erase the word buffer & preceding count
gwlp	byte	XCALL,xKEY,DUP,EMIT
 	byte	XCALL,xdoCHAR
 	byte	_UNTIL,@gw1-@gwlp			'continue building the next word
gw1	byte	EXIT



' ( src -- atrptr | false ) Try to find the string in the dictionary using CMPSTR to help
FINDSTR	' NOTE: tests need to be made on the atr byte to determine code field size
 	byte	PUSH2,(@dictionary+$10)>>8,@dictionary+$10		' from dictionary start
fstlp 	byte	CMPSTR  				' ( src dict+ flg )
 	byte	_IF,@nxtword-@fst1	' found it
fst1	' ( src dict )
 	byte	NIP,EXIT				' ( atrptr ) found
' Skip the attribute byte and codes and test for end of dictionary (entry = 00)
nxtword	' ( src dict ) advance past atr+codes to try next.  (atr(1),bytecode)
nwlp	byte	CFETCHINC,PUSH1,$80,_AND,_UNTIL,@nw1-@nwlp
nw1
 	byte	INC,INC,DUP,CFETCH,ZEQ,_UNTIL,@fst2-@fstlp
fst2	byte	DROP2,_FALSE,EXIT


MAIN	{ apply this label to the main startup word }

' NEW - parse and evaluate input as it is received.
TERMINAL
 	byte	_0FF,XCALL,xms
 	byte	_1,REG,flags,CSTORE
	byte	BL,REG,delim,CSTORE,XCALL,xHEX
	byte  PUSH2,(@rxbufptr+$10)>>8,@rxbufptr+$10,REG,rxbuf,WSTORE
 	byte	_0,REG,CHARS,CSTORE		' init debug virtual input
 	byte	XCALL,xVER
	byte	PRTSTR,"_",0			' subdued prompt
trlp	byte	XCALL,xGETWORD				'Read a character form input stream etc
	byte	REG,wordbuf,XCALL,xFINDSTR
	byte	QDUP,_UNTIL,@qt2-@trlp
qt2	byte	INC,CFETCHINC,SWAP,CFETCH,XCALL,xEXECUTE		' Fetch and execute code (just for testing)
	'byte	XCALL,xDEBUG
	byte	_AGAIN,@qt1-@trlp
qt1


VER 	byte  PRTSTR,$0D,$0A,$0D,$0A," Propeller TACHYON Forth V1.0 rev120718.1100 ",$0D,$0A,0,EXIT



{ DICTIONARY in EEPROM }
{
Although this dictionary is loaded into RAM automatically by the Prop bootloader it is not used by TACHYON
and is free to be used for other purposes. Instead, the copy of the dictionary that is in EEPROM is searched
and this is also where new names are appended to.
Revision: As the image of the cog program's DAT section in RAM can be reused after boot then this would be a good place
to copy the kernel's dictionary from where it is in RAM so it can reuse up some 2KB. Now the kernel's dictionary can be searched in RAM rather than EEPROM! There should be enough room left for another 200 or so entries.

Search methods:

NOTE: The structure of the dictionary has not yet been finalised (undergoing tests)

Structure:
1- Name string
2- Attribute byte (8th bit set also terminates name string )
3- 1st bytecode, 2nd bytecode

Dictionary entries do not need a link field as they are bunched together one after another and it is very easy
to find the next entry by scanning forwards and looking for the attribute byte which will have the msb set then
jumping 3 bytes. A name field that begins with a null indicates end of dictionary (or link to another)
}


CON

' Dictionary header attribute flags
hd	= |<7		' indicates this is a an attribute (delimits the start of a null terminated name)
co	= |<6		'lexicon compile only bit
im	= |<5		'lexicon immediate bit
ex	= |<4		'exec
sm	= |<3		'
rl	= |<2

' code attributes	00 = single bytecode, 01 = relative call, 02 = XCALL bytecode, 03 = WCALL bytecode
xc	= |<1		'XCALL bytecode
rc	= |<0		'RCALL bytecode
ac	= xc+rc	'WCALL - 2 byte address

DAT
'	byte	0[$6C00]		' Place the dictionary far up the end of the EEPROM

{ This is an 8th bit terminated string using the attribute byte so it saves one byte per entry plus it may simplfy the string compare function. Searching still proceeds from lower memory to higher memory and if the atr }



dictionary

	'	NAME,ATR,CODE
 	byte	"RESET",		hd,	RESET,EXIT
 	byte	"STOP",		hd,	STOP,EXIT
	byte	"?EXIT",		hd,	IFEXIT,EXIT
	byte	"0EXIT",		hd,	ZEXIT,EXIT
	byte	"EXIT",		hd,	EXIT,EXIT
 	byte	"NOP",		hd,	_NOP,EXIT
	byte	"3DROP",		hd,	DROP3,EXIT
	byte	"2DROP",		hd,	DROP2,EXIT
  	byte	"DROP",		hd,	DROP,EXIT
 	byte	"?DUP",		hd,	QDUP,EXIT
	byte	"DUP",		hd,	DUP,EXIT
	byte	"2DUP",		hd,	DUP2,EXIT
	byte	"OVER",		hd,	OVER,EXIT
	byte	"3RD",		hd,	THIRD,EXIT
	byte	"4TH",		hd,	FOURTH,EXIT
	byte	"SWAP",		hd,	SWAP,EXIT
	byte	"ROT",		hd,	ROT,EXIT
	byte	"NIP",		hd,	NIP,EXIT


	byte	"1+",		hd,	INC,EXIT
	byte	"1-",		hd,	DEC,EXIT
	byte	"+",		hd,	PLUS,EXIT
	byte	"-",		hd,	MINUS,EXIT
	byte	"*",		hd,	MULTIPLY,EXIT
	byte	"/",		hd,	DIVIDE,EXIT
	byte	"UM*",		hd,	UMMUL,EXIT
	byte	"NEGATE",	hd,	NEGATE,EXIT

	byte	"NOT",		hd,	_NOT,EXIT
	byte	"AND",		hd,	_AND,EXIT
	byte	"OR",		hd,	_OR,EXIT
	byte	"XOR",		hd,	_XOR,EXIT

	byte	"SHR",		hd,	_SHR,EXIT
	byte	"SHL",		hd,	_SHL,EXIT
	byte	"2/",		hd,	_SHR1,EXIT
	byte	"2*",		hd,	_SHL1,EXIT
	byte	"REV",		hd,	_REV,EXIT
	byte	"MASK",		hd,	MASK,EXIT

	byte	"0=",		hd,	ZEQ,EXIT
	byte	"=",		hd,	EQ,EXIT
	byte	">",		hd,	GT,EXIT

	byte	"C@",		hd,	CFETCH,EXIT
	byte	"W@",		hd,	WFETCH,EXIT
	byte	"@",		hd,	FETCH,EXIT
	byte	"C+!",		hd,	CPLUSST,EXIT
	byte	"C!",		hd,	CSTORE,EXIT
	byte	"C@++",		hd,	CFETCHINC,EXIT
	byte	"W+!",		hd,	WPLUSST,EXIT
	byte	"W!",		hd,	WSTORE,EXIT
	byte	"+!",		hd,	PLUSST,EXIT
	byte	"!",		hd,	STORE,EXIT
	byte	"CMOVE",		hd,	CMOVE,EXIT
	byte	"SET",		hd,	SET,EXIT
	byte	"CLR",		hd,	CLR,EXIT
	byte	"BIT?",		hd,	BTEST,EXIT

	byte	"PUSH4",		hd,	PUSH4,EXIT
	byte	"PUSH3",		hd,	PUSH3,EXIT
	byte	"PUSH2",		hd,	PUSH2,EXIT
	byte	"PUSH1",		hd,	PUSH1,EXIT
	byte	"LVAR",		hd,	VARL,EXIT
	byte	"BVAR",		hd,	VARB,EXIT

	byte	"FALSE",		hd,	_0,EXIT
	byte	"OFF",		hd,	_0,EXIT
	byte	"0",		hd,	_0,EXIT
	byte	"1",		hd,	_1,EXIT
	byte	"2",		hd,	_2,EXIT
	byte	"3",		hd,	_3,EXIT
	byte	"4",		hd,	_4,EXIT
	byte	"8",		hd,	_8,EXIT
	byte	"0F",		hd,	_0F,EXIT
	byte	"0FF",		hd,	_0FF,EXIT
	byte	"ON",		hd,	MINUS1,EXIT
	byte	"TRUE",		hd,	MINUS1,EXIT
	byte	"-1",		hd,	MINUS1,EXIT
	byte	"BL",		hd,	BL,EXIT



	byte	"LSTACK",	hd,	LSTACK,EXIT
	byte	"EMIT",		hd,	EMIT,EXIT



	byte	"P@",		hd,	PFETCH,EXIT
	byte	"P!",		hd,	PSTORE,EXIT
	byte	"OUTSET",	hd,	OUTSET,EXIT
	byte	"OUTCLR",	hd,	OUTCLR,EXIT
	byte	"OUTPUTS",	hd,	OUTPUTS,EXIT
	byte	"INPUTS",	hd,	INPUTS,EXIT
	byte	"PX",		hd,	PX,EXIT
	byte	"CLKBIT",	hd,	CLKBIT,EXIT
	byte	"CLKBITS",	hd,	CLKBITS,EXIT
	byte	"COG@",		hd,	COGFETCH,EXIT
	byte	"COG!",		hd,	COGSTORE,EXIT
	byte	"HUBOP",		hd,	_HUBOP,EXIT
	byte	"P@",		hd,	PFETCH,EXIT
	byte	"STACKS",	hd,	STACKS,EXIT
	byte	"LSTACK",	hd,	LSTACK,EXIT

	byte	"XCALL",		hd,	XCALL,EXIT
	byte	"RCALL",		hd,	RCALL,EXIT
	byte	"WCALL",		hd,	WCALL,EXIT
	byte	"CMPJEQ",	hd,	CMPJEQ,EXIT
	byte	"(ELSE)",	hd,	_ELSE,EXIT
	byte	"(IF)",		hd,	_IF,EXIT
	byte	"(UNTIL)",	hd,	_UNTIL,EXIT
	byte	"(AGAIN)",	hd,	_AGAIN,EXIT
	byte	"(ADO)",		hd,	ADO,EXIT
	byte	"(DO)",		hd,	DO,EXIT
	byte	"(LOOP)",	hd,	LOOP,EXIT
	byte	"(+LOOP)",	hd,	PLOOP,EXIT
	byte	"(FOR)",		hd,	FOR,EXIT
	byte	"(NEXT)",	hd,	forNEXT,EXIT

	byte	">R",		hd,	PUSHR,EXIT
	byte	"R>",		hd,	RPOP,EXIT
	byte	">L",		hd,	PUSHL,EXIT
	byte	"L>",		hd,	LPOP,EXIT
	byte	"I",		hd,	I,EXIT

	byte	"BKEY",		hd,	BKEY,EXIT
	byte	"EMIT",		hd,	EMIT,EXIT
	byte	"(REG)",		hd,	REG,EXIT
	byte	"@REG",		hd,	ATREG,EXIT
	byte	"WAIT",		hd,	WAIT,EXIT
'	byte	"WAITCNT",	hd,	_WAITCNT,EXIT
	byte	"(PTRSTR)",	hd,	PRTSTR,EXIT
	byte	"(CMPSTR)",	hd,	CMPSTR,EXIT



	byte	".S",		hd+xc,	XCALL,xPRTSTK
	byte	"DUMP",		hd+xc,	XCALL,xDUMP
	byte	".L",		hd+xc,	XCALL,xPRTLONG


enddict	byte	0,0

last

{
''****************************************************
''*  HIGH-SPEED PRECISION SERIAL RECEIVE             *
''*  HS-SerialRx                                     *
''*  2008,2012 Peter Jakacki                         *
 ''***************************************************

'17/07/12 Modify for use with Tachyon Forth
'Need to supply the buffer address back from start routine perhaps
'Use the cog image in hub RAM for the buffer
'Interface is simple read & write indexes



PUB start(rxdpin, baudrate)
  long[@rxpin] := |<rxdpin
  long[@bitticks] := (clkfreq / baudrate)
  result := @HSSerialRx+4       ' Use cog image for serial buffer (skip 4 for control)
  cognew(@HSSerialRx, 0)

DAT

                        org
HSSerialRx
                        mov     Y0,rxbuf
                        sub     Y0,#4
                        mov     X0,#0
                        wrlong  X0,Y0
                        mov     stticks,bitticks
                        shr     stticks,#1
                        sub     stticks,#10             ' compensate timing
 receive                mov     rxdata,#0
                        waitpne rxpin,rxpin
'
' START BIT DETECTED
'                                                        'time sample for middle of start bit
                        mov     rxcnt,cnt
                        add     rxcnt,stticks           ' special start bit timing
                        waitcnt rxcnt,bitticks
                        'sample middle of start bit
                        test    rxpin,ina       wz       'sample middle of start bit
rxcond2       if_nz     jmp     #receive                ' restart if false start
'
' START bit validated
' Read in data bits
' No point in looping as we have plenty of code to play with
' and inlining can lead to higher receive speeds
'
                        waitcnt rxcnt,bitticks
                        test    rxpin,ina wc
              if_c      or      rxdata,#01
                        waitcnt rxcnt,bitticks
                        test    rxpin,ina wc
              if_c      or      rxdata,#02
                        waitcnt rxcnt,bitticks
                        test    rxpin,ina wc
              if_c      or      rxdata,#04
                        waitcnt rxcnt,bitticks
                        test    rxpin,ina wc
              if_c      or      rxdata,#08
                        waitcnt rxcnt,bitticks
                        test    rxpin,ina wc
              if_c      or      rxdata,#$10
                        waitcnt rxcnt,bitticks
                        test    rxpin,ina wc
              if_c      or      rxdata,#$20
                        waitcnt rxcnt,bitticks
                        test    rxpin,ina wc
              if_c      or      rxdata,#$40
                        waitcnt rxcnt,bitticks
                        test    rxpin,ina wc
              if_c      or      rxdata,#$80
                        waitcnt rxcnt,bitticks          ' check stop bit
                        test    rxpin,ina wc
        if_nc           jmp     #receive                  ' discard if framing error (no stop bits)
wrbuf                   rdbyte  Y0,rxwr         ' rxwr points to byte index in hub
                        mov     X0,rxbuf
                        add     X0,Y0             ' X points to buffer location to store
                        wrbyte  rxdata,X0
                        add     Y0,#1
                        wrbyte  Y0,rxwr
                        jmp     #receive                'byte done, receive next byte

                        long    0[16]           ' make sure this code image is more than 256+4 bytes


rxpin                   long    0               'mask of rx pin
rxwr                    long    @@@HSSerialRx+3    ' byte address of rxwr in hub
rxbuf                   long    @@@HSSerialRx+4    'pointer to rxbuf in hub memory

bitticks                long    0
stticks                 long    0
rxcnt                   res     1
rxdata                  res     1               'assembled character
X0                      res     1
Y0                      res     1

}
